package asd.td4;

import asd.base.Liste;
import java.lang.Math;

/**
 * Une classe pour l'exercice 5 de la feuille 4.
 * Lorsqu'elle n'est pas mentionne, la complexit est O(1).
 *  2002 Marc Gatano
 */
public class Polynome {

    // CLASSE INTERNE

    /**
     * Une classe pour reprsenter des monmes. On dit qu'un monme a.X^n
     * est plus "grand" qu'un monme b.X^m si n > m.
     */
    private class Monome {

        private int expo; // l'exposant du monme
        private int coef; // le coefficient du monme

        /**
         * Construit le monme de coefficient coef et
         * d'exposant expo.
         * Prcondition : expo >= 0.
         */
        public Monome(int coef, int expo) {
            this.coef = coef;
            this.expo = expo;
        }

        /**
         * Teste si le monme est nul.
         */
        public boolean nul() {
            return coef == 0;
        }

        /**
         * Retourne le monme somme du monme et du monme m.
         * Prcondition : this.expo = m.expo.
         */
        public Monome plus(Monome m) {
            return new Monome(coef + m.coef,expo);
        }

        /**
         * Retourne le monme produit du monme et du monme m.
         */
        public Monome fois(Monome m) {
            return new Monome(coef * m.coef,expo + m.expo);
        }

        /**
         * Teste si this.expo < m.expo.
         */
        public boolean inferieur(Monome m) {
            return expo < m.expo;
        }

        /**
         * Evalue le monme en i, autrement dit calcule la
         * valeur de c * (e^i) o c et e sont respectivement
         * le coefficient et l'exposant du monme.
         */
        public int evaluer(int i) {
            return coef * (int) Math.pow((double) i,(double) expo);
        }

        public boolean egal(Monome m) {
            return coef == m.coef && expo == m.expo;
        }

        /**
         * Retourne une chane de caractres correspondant  la
         * reprsentation externe du monme vu comme le premier
         * monme d'un polynme.
         */
        public String toString1() {
            String s = "";
            if ( coef < 0 ) {
                s = " - ";
            }
            if ( Math.abs(coef) > 1 || expo == 0 ) {
                s = s + String.valueOf(Math.abs(coef));
            }
            if ( expo > 0 ) {
                s = s + "X";
            }
            if ( expo > 1 ) {
                s = s + "^" + String.valueOf(expo);
            }
            return s;
        }

        /**
         * Retourne une chane de caractres correspondant  la
         * reprsentation externe du monme vu comme un monme
         * d'un polynme quand ce monme n'est pas le premier.
         */
        public String toString() {
            if ( coef > 0 ) {
                return " + " + toString1();
            }
            return toString1();
        }
    }

    // ATTRIBUTS

    private Liste l; // la liste strictement dcroissante (sur
                           // les exposants) des monmes
    // CONSTRUCTEURS

    /**
     * Construit le polynme nul.
     */
    public Polynome() {
        l = new Liste();
    }

    /**
     * Construit le polynme rduit au monme a.X^n avec
     * a = coef et n = expo.
     * Prcondition : expo >= 0.
     */
    public Polynome(int coef, int expo) {
        if ( coef != 0 ) {
            l = new Liste(new Monome(coef,expo));
        }
        else {
            l = new Liste();
        }
    }

    /**
     * Construit un polynme dont le plus grand monme est m
     * et les autres monme sont ceux du polynme p.
     * Prcondition : pour tout monme c.X^k de p, k < n
     * o n est l'exposant de m.
     */
    private Polynome(Monome m, Polynome p) {
        l = new Liste(m,p.l);
    }

    /**
     * Construit un polynme  partir de la liste l.
     * Prcondition : la liste l est une liste strictement dcroissante
     * de monmes.
     */
    private Polynome(Liste l) {
        this.l = l;
    }

    /**
     * Construit le polynme rduit au monme m.
     * Prcondition : m n'est pas nul.
     */
    private Polynome(Monome m) {
        l = new Liste(m);
    }

    // PRIMITIVES

    /**
     * Retourne le plus grand monme du polynme.
     */
    private Monome premier() {
        return (Monome) l.premier();
    }

    /**
     * Retourne le polynme priv de son plus grand monme.
     */
    private Polynome reste() {
        return new Polynome(l.reste());
    }

    /**
     * Teste si le polynme est nul.
     */
    public boolean nul() {
        return l.vide();
    }

    // NON-PRIMITIVES  CRIRE  l'AIDE DES PRIMITIVES

    /**
     * Retourne la somme du polynme avec le polynme p.
     * Complexit : O(max(t,tp)) o t et tp sont respectivement
     * les nombres de termes de this et de p.
     */
    public Polynome plus(Polynome p) {
        if ( nul() ) {
            return p;
        }
        if ( p.nul() ) {
            return this;
        }
        if ( premier().inferieur(p.premier()) ) {
            return new Polynome(p.premier(),plus(p.reste()));
        }
        if ( p.premier().inferieur(premier()) ) {
            return new Polynome(premier(),reste().plus(p));
        }
        Monome m = premier().plus(p.premier());
        if ( m.nul() ) {
            return reste().plus(p.reste());
        }
        return new Polynome(m,reste().plus(p.reste()));
    }

    /**
     * Retourne la somme du polynme avec l'entier i.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    public Polynome plus(int i) {
        return plus(new Polynome(new Monome(i,0)));
    }

    /**
     * Retourne la diffrence entre le polynme et le polynme p.
     * Complexit : O(max(t,tp)) o t et tp sont respectivement
     * les nombres de termes de this et de p.
     */
    public Polynome moins(Polynome p) {
        return plus(p.fois(-1));
    }

    /**
     * Retourne la diffrence entre le polynme et l'entier i.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    public Polynome moins(int i) {
        return plus(new Polynome(new Monome(- i,0)));
    }

    /**
     * Retourne le produit du polynme par le polynme p.
     * Complexit : O(t*tp) o t et tp sont respectivement
     * les nombres de termes de this et de p.
     */
    public Polynome fois(Polynome p) {
        if ( nul() || p.nul() ) {
            return new Polynome();
        }
        return p.fois(premier()).plus(reste().fois(p));
    }

    /**
     * Retourne le produit du polynme par le monme m.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    private Polynome fois(Monome m) {
        if ( nul() ) {
            return new Polynome();
        }
        return new Polynome(premier().fois(m),reste().fois(m));
    }

    /**
     * Retourne le produit du polynme par l'entier i.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    public Polynome fois(int i) {
        return fois(new Monome(i,0));
    }

    /**
     * Teste si le polynme est gal au polynme p.
     * Complexit : O(min(t,tp)) o t et tp sont respectivement
     * les nombres de termes de this et de p.
     */
    public boolean egal(Polynome p) {
        if ( nul() ) {
            return p.nul();
        }
        if ( p.nul() ) {
            return false;
        }
        return premier().egal(p.premier()) && reste().egal(p.reste());
    }

    /**
     * Retourne la valeur du polynme pour X = i.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    public int evaluer(int i) {
        if ( nul() ) {
            return 0;
        }
        return premier().evaluer(i) + reste().evaluer(i);
    }

    /**
     * Retourne une chane de caractres reprsentant le polynme.
     * Complexit : O(t) o t est le nombre de termes de this.
     */
    public String toString() {
        String s = "";
        if ( nul() ) {
            return "0";
        }
        s = s + premier().toString1();
        Polynome r = reste();
        while ( ! r.nul() ) {
            s = s + r.premier().toString();
            r = r.reste();
        }
        return s;
    }
}
